#include <CoreFoundation/CoreFoundation.h>
#include <syslog.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/IOBSD.h>
#include <IOKit/storage/ata/IOATAStorageDefines.h>

io_object_t getATADeviceForBSDPath(const char *path);
void printDeviceID(io_object_t disk);
int setAPMLevel(io_object_t disk, unsigned char level);
void usage();

static bool log_to_syslog = false;

#define HDAPM_VERSION "1.2"

void log_info( const char* format, ... ) {
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
  if (log_to_syslog)
  {
    va_start(args, format);
    vsyslog(LOG_NOTICE, format, args);
    va_end(args);
  }
}

int main (int argc, const char * argv[]) {
  unsigned char apm_level = 0;
  
  
  if (argc < 2)
	usage();

  const char **devices = malloc(sizeof(char *) * argc);
  int devCount = 0;
  
  for (int i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "--help") == 0)
      usage();
    else if (strcmp(argv[i], "--version") == 0)
    {
      printf("hdapm version %s\n", HDAPM_VERSION);
      exit(0);
    }
    else if (strcmp(argv[i], "-log") == 0)
    {
      log_to_syslog = true;
      openlog("hdapm", 0, 0);
    }
    /* Check for a device ID. Assumes they start with 'disk'. */
    else if (strncmp(argv[i], "disk", 4) == 0)
      devices[devCount++] = argv[i];
    /* Parse APM Level */
    else if (strcmp(argv[i], "min") == 0)
      apm_level = kIOATAMaxPowerSavings;
    else if (strcmp(argv[i], "default") == 0)
      apm_level = kIOATADefaultPerformance;
    else if (strcmp(argv[i], "max") == 0)
      apm_level = kIOATAMaxPerformance;
    else /* Numeric value */
    {
      long l = strtoul(argv[i], NULL, 0);
      if (l < 0x01 || l > 0xfe)
      {
        if (l == 0 && errno == EINVAL)
          log_info("Error: Unknown argument '%s'\n", argv[i]);
        else
          log_info("Error: LEVEL must be between 1 and 254.\n");
        exit(1);
      }
      else
        apm_level = (unsigned char) l;
    }
  }
  if (apm_level == 0)
    usage();
  
  int r = 0;
  
  if (devCount == 0)
  {
    /* User didn't specify any device IDs. Look up all ATA devices. */
    CFMutableDictionaryRef matcher = IOServiceMatching("IOAHCIBlockStorageDevice");
    io_iterator_t iterator;
    IOServiceGetMatchingServices(kIOMasterPortDefault, matcher, &iterator);
    
    io_object_t hdd;
    while((hdd = IOIteratorNext(iterator)))
    {
      printDeviceID(hdd);
      r &= setAPMLevel(hdd, apm_level);
      IOObjectRelease(hdd);
    }
  }
  else /* devCount > 0 */
  {
    for (int i=0; i < devCount; i++)
    {
      io_object_t hdd = getATADeviceForBSDPath(devices[i]);
      if (!hdd)
        continue;
      printDeviceID(hdd);
      r &= setAPMLevel(hdd, apm_level);
      IOObjectRelease(hdd);
    }
  }
  free(devices);
  if (log_to_syslog)
    closelog();
  return r;
}

void usage()
{
  fprintf(stderr, "Usage: hdapm [OPTION]... [DEVICE]... LEVEL\n\
Set ATA Advanced Power Management level.\n\
\n\
Example: hdapm max\n\
         hdapm disk0 max\n\
\n\
DEVICE	 (Optional) BSD path of ATA device\n\
LEVEL	 numeric value 1 - 254, or...\n\
		 min (best power savings)\n\
		 max (maximum performance)\n\
		 default\n\
OPTIONS\n\
  -log           Write status to system log\n\
      --help     display this help and exit\n\
      --version  output version information and exit\n");
  exit(1);
}

int setAPMLevel(io_object_t disk, unsigned char level)
{
  int l = (int) level;
  
  CFNumberRef cfLevel = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &l);
  kern_return_t r = IORegistryEntrySetCFProperty(disk, CFSTR("APM Level"), cfLevel);
  CFRelease(cfLevel);
  
  const char *result;
  if (r)
  {
	if (r == kIOReturnUnsupported)
	  result = "FAILED: APM not supported";
	else if (r == kIOReturnNotPrivileged)
      result = "FAILED: Permission denied";
    else
	  result = "FAILED";
  }
  else
    result = "Success";
  
  log_info("  Set APM Level to 0x%02x: %s\n", level, result);
  
  return r;
}

bool is_ata_disk(io_object_t device)
{
  /* In OS X <= 10.4, the driver to look for is called IOATABlockStorage.
    In 10.5, it is called IOAHCIBlockStorage.  */
  if (IOObjectConformsTo(device, "IOAHCIBlockStorageDevice")
	  || IOObjectConformsTo(device, "IOATABlockStorageDevice"))
	return TRUE;
  return FALSE;
}

io_object_t find_ata_device(io_object_t device)
{
  io_object_t child = 0;
  IOReturn r;
  while (! is_ata_disk(device))
  {
	child = device;
	r = IORegistryEntryGetParentEntry(child, kIOServicePlane, &device);
	IOObjectRelease(child);
	if (r != 0 || device == 0)
	  return 0;
  }
  return device;
}

void printDeviceID(io_object_t disk)
{
  CFStringRef bsdName = IORegistryEntrySearchCFProperty(disk, kIOServicePlane, CFSTR(kIOBSDNameKey), NULL, kIORegistryIterateRecursively);
  const char *bsd_str = CFStringGetCStringPtr(bsdName, kCFStringEncodingMacRoman);
  
  CFTypeRef dict = IORegistryEntryCreateCFProperty(disk, CFSTR("Device Characteristics"), NULL, 0);
  if (dict)
  {
	CFStringRef str = CFDictionaryGetValue(dict, CFSTR("Product Name"));
	if (str)
	{
	  char devID[100];
	  CFStringGetCString(str, devID, 100, kCFStringEncodingUTF8);
	  log_info("%s: %s\n", bsd_str, devID);
	}
	CFRelease(dict);	  
  }
  CFRelease(bsdName);
}

io_object_t getATADeviceForBSDPath(const char* path)
{
  const char *device_name;

  /* IOBSDNameMatching doesn't want the leading path. */
  if (strncmp (path, "/dev/", 5) == 0)
	device_name = path + 5;
  else
	device_name = path;
  
  CFMutableDictionaryRef matcher = IOBSDNameMatching(kIOMasterPortDefault, 0, device_name);
  io_object_t device = IOServiceGetMatchingService(kIOMasterPortDefault, matcher);
  if (! device)
  {
    log_info("%s: not found\n", path);
    return 0;
  }

  device = find_ata_device(device);
  if (! device)
  {
	log_info("%s: not an ATA device\n", path);
    return 0;
  }
  return device;
}

